{*****************************************************************************}
{                                                                             }
{    Tnt Extended Editors                                                     }
{      http://lummie.co.uk                                                    }
{        Version: 1.0.3                                                       }
{                                                                             }
{    Copyright (c) 2005 Matt harrison (http://www.lummie.co.uk)               }
{    Descended from TNT Controls                                              }
{        Copyright (c) 2002, 2003 Troy Wolbrink (troy.wolbrink@ccci.org)      }
{                                                                             }
{*****************************************************************************}

{
  1.0.3
  15-9-2005 - Added PostShowMessage
}

unit TntExDropDownEdit;

interface

uses
  SysUtils, Classes, Controls, StdCtrls, TntStdCtrls, TntExButtonEdit, windows, messages,
  ExtCtrls, graphics, TntMenus;

const
  CM_ChangePanelVisiblity = WM_USER + 875;
  ResizeSize = 4;

type
  TCMChangePanelVisible = TWMShowWindow;

  TCustomTntExDropDownEdit = class;

  TTntExDropDownPanel = class(TCustomPanel)
  protected
    FDropDown : TCustomTntExDropDownEdit;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure MouseMove(Shift: TShiftState; X: Integer; Y: Integer);
      override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X: Integer; Y: Integer); override;
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CreateWnd; override;
    procedure WMActivateApp(var Message: TWMActivateApp);
      message WM_ACTIVATEAPP;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    Procedure PostShowMessage(IsVisible : boolean);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TTntExDropDownStyle = (csDropDown, csDropDownList);

  TCustomTntExDropDownEdit = class(TTntEdit)
  private
    { Private declarations }
    FPanel : TTntExDropDownPanel;
    FPopulated : boolean;
    FPanelHeight: integer;
    FPanelWidth: integer;
    FClosing : boolean;
    FOnDroppedDown : TNotifyEvent;
    FOnClosedUp: TNotifyEvent;
    FCanvas : TControlCanvas;
    FButtonPressed : boolean;
    FMouseInControl : boolean;
    FUpdating : boolean;
    FDropDownPopupMenu: TTntPopupMenu;
    FDropDownStyle: TTntExDropDownStyle;
    FRedirectingContextMenu : boolean;
    FPanelAutoWidth: boolean;
    FPanelAllowResize : boolean;
    function GetCanvas: TCanvas;
  protected
    { Protected declarations }
    procedure GetPanelRect(var ARect : TRect); virtual;
    procedure DoButtonClick; virtual;
    function GetButtonRect : TRect; virtual;
    function GetButtonWidth : integer; virtual;
    procedure CMChangePanelVisibility(var Msg : TCMChangePanelVisible); message CM_ChangePanelVisiblity;
    procedure CreateWnd; override;
    function DoPopulatePanel(APanel : TTntExDropDownPanel) : boolean; virtual;
    procedure DoDroppedDown; virtual;
    procedure DoClosedUp; virtual;
    procedure UpdateMargins; virtual;
    procedure UpdatePanelPosition; virtual;
    procedure WMMove(var Message: TWMMove); message WM_MOVE;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure Change; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X: Integer; Y: Integer); override;
    function HitButton(x,y : integer) : boolean; virtual;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X: Integer;
      Y: Integer); override;
    procedure CMMouseenter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseleave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure MouseMove(Shift: TShiftState; X: Integer; Y: Integer);
      override;
    procedure DoExit; override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
      override;
    procedure WndProc(var Message: TMessage); override;
    procedure DropDown; dynamic;
    procedure CloseUp; dynamic;
    function GetDropDownStyle : TTntExDropDownStyle; virtual;
    procedure WMLButtonDown(var Message: TWMLButtonDown);
      message WM_LBUTTONDOWN;
    procedure NonEditMouseDown(var Message: TWMLButtonDown);
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure KeyPress(var Key: Char); override;
    property Panel : TTntExDropDownPanel read FPanel;
    procedure DoCustomPaint; virtual;
    property Canvas : TCanvas read GetCanvas;
  public
    { Public declarations }
    procedure ShowPanel(MakeVisible: boolean);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Closing : boolean read FClosing;
    function IsDroppedDown : boolean;
    function IsEditable : boolean; 
    property PanelAutoWidth : boolean read FPanelAutoWidth write FPanelAutoWidth;
    property PanelWidth : integer read FPanelWidth write FPanelWidth;
    property PanelHeight : integer read FPanelHeight write FPanelHeight;
    property PanelAllowResize : boolean read FPanelAllowResize write FPanelAllowResize;
    property OnDroppedDown : TNotifyEvent read FOnDroppedDown write FOnDroppedDown;
    property OnClosedUp : TNotifyEvent read FOnClosedUp write FOnClosedUp;
    property DropDownPopupMenu : TTntPopupMenu read FDropDownPopupMenu write FDropDownPopupMenu;
    property DropDownStyle : TTntExDropDownStyle read FDropDownStyle write FDropDownStyle;
  end;

  TTntExDropDownEdit = class(TCustomTntExDropDownEdit)
  published
    { Published declarations }
    property PanelAutoWidth;
    property PanelWidth;
    property PanelHeight;
    property PanelAllowResize;
    property OnDroppedDown;
    property OnClosedUp;
    property DropDownPopupMenu;
    property DropDownStyle;
  end;

implementation

uses
  themes;

{ TTntExDropDownPanel }

procedure TTntExDropDownPanel.AdjustClientRect(var Rect: TRect);
begin
  Rect.bottom := Rect.bottom - ResizeSize;
  inherited;
end;

constructor TTntExDropDownPanel.Create(AOwner: TComponent);
begin
  inherited;
  FDropDown := TCustomTntExDropDownEdit(AOwner);
  ControlStyle := ControlStyle + [csNoDesignVisible] - [csParentBackground];
  visible := false;
  TabStop := false;
  Constraints.MinHeight := 10;
  Constraints.MinWidth := 10;
end;

procedure TTntExDropDownPanel.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or WS_POPUP;
    ExStyle := ExStyle or WS_EX_TOOLWINDOW + WS_EX_LEFT  and not WS_EX_APPWINDOW;
    AddBiDiModeExStyle(ExStyle);
    WindowClass.Style := CS_SAVEBITS;
  end;
end;


procedure TTntExDropDownPanel.CreateWnd;
begin
  inherited;
  if not (csDesigning in ComponentState) then
    Windows.SetParent(Handle, 0);
  CallWindowProc(DefWndProc, Handle, WM_SETFOCUS, 0, 0);
end;

destructor TTntExDropDownPanel.Destroy;
begin
  inherited;
end;

procedure TTntExDropDownPanel.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if (Key = VK_F4) then
  begin
    PostShowMessage(false);
  end;
end;

procedure TTntExDropDownPanel.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
const
  SC_DRAGSIZE_SE = $F008;
  SC_DRAGSIZE_S = $F006;
begin
  if FDropDown.PanelAllowResize then
  begin
    if ptinRect(Rect(0, height-ResizeSize, width-ResizeSize, height),point(x,y)) then
    begin
      ReleaseCapture;
      FDropDown.PanelAutoWidth := false;
      Perform(WM_SYSCOMMAND,SC_DRAGSIZE_S,0);
    end
    else if ptinRect(Rect(width-ResizeSize,height-ResizeSize,width,height),point(x,y)) then
    begin
      ReleaseCapture;
      FDropDown.PanelAutoWidth := false;
      Perform(WM_SYSCOMMAND,SC_DRAGSIZE_SE,0);
    end
    else
      inherited;
  end
  else
  begin
    inherited;
  end;
end;

procedure TTntExDropDownPanel.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if FDropDown.PanelAllowResize then
  begin
    if ptinRect(Rect(0, height-ResizeSize, width-ResizeSize, height),point(x,y)) then
      cursor := crSizeNS
    else if ptinRect(Rect(width-ResizeSize,height-ResizeSize,width,height),point(x,y)) then
      cursor := crSizeNWSE
    else
      cursor := crDefault;
  end;
end;

procedure TTntExDropDownPanel.PostShowMessage(IsVisible : boolean);
begin
  if assigned(FDropDown) and FDropDown.HandleAllocated then
    PostMessage(FDropdown.Handle,CM_ChangePanelVisiblity,integer(isVisible),0);
end;

procedure TTntExDropDownPanel.WMActivate(var Message: TWMActivate);
begin
  if (Message.Active = WA_INACTIVE) and not(FDropDown.Closing) then
  begin
    if assigned(FDropDown) then FDropDown.ShowPanel(false)
  end
  else
    inherited;
end;

{ TTntDropDownEdit }

procedure TCustomTntExDropDownEdit.CMChangePanelVisibility(
  var Msg: TCMChangePanelVisible);
begin
  // override sub controls can post a CM_ChangePanelVisibility message to hide and show the panel 
  ShowPanel(msg.Show);
end;

constructor TCustomTntExDropDownEdit.Create(AOwner: TComponent);
begin
  inherited;
  FPanelAllowResize := true;
  FPanelAutoWidth := true;
  FDropDownPopupMenu := nil;
  FUpdating := false;
  FButtonPressed := false;
  FCanvas := nil;
  FPopulated := false;
  FPanel := TTntExDropDownPanel.create(self);
  FPanel.parent := self;
  FPanel.Visible := false;
  FClosing := false;
  FOnDroppedDown := nil;
  FOnClosedUp := nil;
  FPanelHeight := 0;
  FPanelWidth := 0;
  FDropDownStyle := GetDropDownStyle;
end;

procedure TCustomTntExDropDownEdit.CreateWnd;
begin
  inherited;
  FPanel.HandleNeeded;
  UpdateMargins;
end;

destructor TCustomTntExDropDownEdit.Destroy;
begin
  FreeAndNil(FCanvas);
  inherited;
end;

procedure TCustomTntExDropDownEdit.ShowPanel(MakeVisible : boolean);
begin
  if MakeVisible then DropDown else CloseUp;
end;

procedure TCustomTntExDropDownEdit.GetPanelRect(var ARect: TRect);
begin
  // returns where to display the drop down.
  // note the right and bottom are width and height
  if (FPanelHeight = 0) and (FPanelWidth = 0) then
  begin
    FPanelHeight := Height * 8;
    FPanelWidth := width + GetButtonWidth - 1;
  end;

  ARect.TopLeft := ClientOrigin;
  ARect.Left := ARect.left -2;
  ARect.Top := ARect.Top + Height -2;
  if FPanelAutoWidth then
    ARect.Right := Width
  else
    ARect.Right := FPanelWidth;
  ARect.bottom := FPanelHeight;
end;


function TCustomTntExDropDownEdit.DoPopulatePanel(
  APanel: TTntExDropDownPanel): boolean;
begin
  // override to create a sub control of the panel, e.g. a list box and do the initiali setup here.
  // if you return true this function will never get called again
  // if you return false, this function will get called just before everytime the panel is dropped down.
  result := true;
end;

procedure TCustomTntExDropDownEdit.DoClosedUp;
begin
  // override to dow something after the panel is closed up
  if assigned(FOnClosedUp) then FOnClosedUp(self);
end;

procedure TCustomTntExDropDownEdit.DoDroppedDown;
begin
  // override to do something after the panel is droped down
  if assigned(FOnDroppedDown) then FOnDroppedDown(self);
end;


procedure TCustomTntExDropDownEdit.UpdateMargins;
begin
  Perform(EM_SETMARGINS,EC_RIGHTMARGIN,(GetButtonWidth+4)*$10000);
end;

procedure TCustomTntExDropDownEdit.UpdatePanelPosition;
var
  ARect : TRect;
begin
  // this procedure actuall moves the panel to the correct position
  // override getPanelRect to move it elsewhere
  GetPanelRect(ARect);
  if assigned(FPanel) then FPanel.SetBounds(ARect.Left,ARect.Top,ARect.Right,ARect.Bottom);
  SetWindowPos(FPanel.Handle,0,ARect.Left,ARect.Top,ARect.Right,ARect.Bottom, SWP_NOACTIVATE);
end;

procedure TCustomTntExDropDownEdit.WMMove(var Message: TWMMove);
begin
  UpdatePanelPosition;
  UpdateMargins;
end;

procedure TCustomTntExDropDownEdit.WMSize(var Message: TWMSize);
begin
  UpdateMargins;
end;

procedure TCustomTntExDropDownEdit.Change;
begin
  inherited;
  UpdateMargins;
end;

procedure TCustomTntExDropDownEdit.DoButtonClick;
begin
  // override to alter what happens when the button is clicked
  SetFocus;
  if not FClosing then ShowPanel(not FPanel.Visible);
end;

procedure TCustomTntExDropDownEdit.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if HitButton(x,y) then
  begin
    FButtonPressed := true;
    DoButtonClick;
  end;
  invalidate;
end;

function TCustomTntExDropDownEdit.HitButton(x, y: integer): boolean;
var
  pt : TPoint;
begin
  // override to change the button hit test if needed
  pt := point(x,y);
  result := ptinrect(Rect(Width-GetButtonWidth,0,width,height),pt);
end;

function TCustomTntExDropDownEdit.GetButtonWidth: integer;
begin
  // override to change the width of the button
  result := GetSystemMetrics(SM_CXVSCROLL) + 2;
end;

procedure TCustomTntExDropDownEdit.WMPaint(var Message: TWMPaint);
var
  DC, oldDC : HDC;
  PS : TPaintStruct;
  State : TThemedComboBox;
  Details: TThemedElementDetails;
  R : TRect;
  Flags : integer;
begin
  if csPaintCopy in ControlState then
  begin
    inherited;
    exit;
  end;
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;

  if ThemeServices.ThemesEnabled then
  begin
    R := GetButtonRect;
    DC := Message.DC;
    if DC = 0 then DC := BeginPaint(Handle, PS);
    FCanvas.Handle := DC;
    try
      OldDC := Message.DC;
      Message.DC := DC;
      inherited;
      Message.DC := oldDC;
      if not enabled then
        State := tcDropDownButtonDisabled
      else
        if FButtonPressed then
          State := tcDropDownButtonPressed
        else
          if FMouseInControl and not FPanel.Visible then
            State := tcDropDownButtonHot
          else
            State := tcDropDownButtonNormal;
      Details := ThemeServices.GetElementDetails(State);
      r.Bottom := r.Bottom ;
      ThemeServices.DrawElement(FCanvas.Handle, Details, R);
      DoCustomPaint;
    finally
      FCanvas.Handle := 0;
      if Message.DC = 0 then EndPaint(Handle, PS);
    end;
  end

  else

  begin
    // NON THEMED DRAWING
    R := GetButtonRect;

    DC := Message.DC;
    if DC = 0 then DC := BeginPaint(Handle, PS);
    FCanvas.Handle := DC;
    try
      Message.DC := DC;
      inherited PaintTo(FCanvas,-2,-2);

      if not (Enabled) then
        Flags := DFCS_SCROLLCOMBOBOX or DFCS_INACTIVE
      else if FButtonPressed then
        Flags := DFCS_SCROLLCOMBOBOX or DFCS_FLAT or DFCS_PUSHED
      else
        Flags := DFCS_SCROLLCOMBOBOX;
      r.Left := r.left -1;
      r.right := r.right -1;
      r.bottom := r.bottom -1;
      DrawFrameControl(FCanvas.Handle, R, DFC_SCROLL, Flags);
      DoCustomPaint;
    finally
      FCanvas.Handle := 0;
      if Message.DC = 0 then EndPaint(Handle, PS);
    end;
  end;


end;

procedure TCustomTntExDropDownEdit.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  FButtonPressed := false;
  invalidate;
end;

procedure TCustomTntExDropDownEdit.CMMouseenter(var Message: TMessage);
begin
  inherited;
  {Windows XP themes use hot track states, hence we have to update the drop down button.}
  if ThemeServices.ThemesEnabled and not (FMouseInControl) and not (csDesigning in ComponentState) then
  begin
    FMouseInControl := True;
    Invalidate;
  end;
end;

procedure TCustomTntExDropDownEdit.CMMouseleave(var Message: TMessage);
begin
  inherited;
  if ThemeServices.ThemesEnabled and FMouseInControl then
  begin
    FMouseInControl := False;
    Invalidate;
    Cursor := crDefault;
  end;
end;

function TCustomTntExDropDownEdit.GetButtonRect: TRect;
begin
  // override to change the position of the button
  result := Rect(Width-GetButtonWidth,-1,width-3,height-3);
end;

procedure TCustomTntExDropDownEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if HitButton(x,y) or not IsEditable then
    Cursor := crArrow
  else
    Cursor := crDefault;
end;

procedure TCustomTntExDropDownEdit.DoExit;
begin
  inherited;
  If IsDroppedDown then ShowPanel(false);
  SelStart := 0;
  SelLength := 0;
  if not IsEditable then Invalidate;
end;

function TCustomTntExDropDownEdit.IsDroppedDown: boolean;
begin
  result := FPanel.Visible;
end;

procedure TCustomTntExDropDownEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited;
end;

procedure TCustomTntExDropDownEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;

  if (Key = VK_Down) and (ssAlt in Shift) then
  begin
    ShowPanel(not IsDroppedDown);
    Key := 0;
  end;

  if (Key = VK_F4) then
  begin
    ShowPanel(not IsDroppedDown);
    Key := 0;
  end;

  if (Key = VK_RETURN) or (key = VK_Escape) then
  begin
    if IsDroppedDown then CloseUp;
    Key := 0;
  end;

  case Key of
    VK_Home, VK_END, VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN : if not IsEditable then Key := 0;
  end;
end;

procedure TCustomTntExDropDownEdit.WMActivate(var Message: TWMActivate);
begin
  if (Message.Active = WA_INACTIVE) and not(Closing) then
    ShowPanel(false)
  else
    inherited;
  UpdateMargins;
end;


procedure TTntExDropDownPanel.WMActivateApp(var Message: TWMActivateApp);
begin
  if (message.Active = false) and (visible) then FDropDown.CloseUp;
end;

procedure TTntExDropDownPanel.WMSize(var Message: TWMSize);
begin
  if assigned(FDropDown) then
  begin
    FDropDown.FPanelWidth := Message.Width;
    FDropDown.FPanelHeight := Message.Height;
  end;
  realign;
end;

procedure TCustomTntExDropDownEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (operation = opRemove) and (AComponent = FDropDownPopupMenu) then FDropDownPopupMenu := nil;
end;

procedure TCustomTntExDropDownEdit.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_CONTEXTMENU :
    begin
      if assigned(FPanel) and (FPanel.HandleAllocated) and (Fpanel.Visible) then
      begin
        if (TWMContextMenu(Message).hWnd = FPanel.handle) then // right click on actual drop down then throw away.
        begin
          message.Result := 0;
        end
        else
        begin  // send message to panel
          if not FRedirectingContextMenu then
          begin
            FRedirectingContextMenu := true;
            with TMessage(Message) do
              SendMessage(FPanel.Handle, Msg, WParam, LParam);
            FRedirectingContextMenu := false;
          end
          else
          begin
            if Parent.HandleAllocated then
              with TMessage(Message) do
                SendMessage(Parent.Handle, Msg, WParam, LParam);
          end;
        end;
      end
      else
      begin
        inherited;
      end;
    end;
  else
    inherited;
  end
end;

procedure TCustomTntExDropDownEdit.CloseUp;
begin
  FClosing := true;
  SetFocus;
  FPanel.Visible := false;
  DoClosedUp;
  FClosing := false;
  if not IsEditable then HideCaret (Handle);
end;

procedure TCustomTntExDropDownEdit.DropDown;
begin
  UpdatePanelPosition;
  if not FPopulated then
  begin
    FPopulated := DoPopulatePanel(FPanel);
  end;
  FPanel.Visible := True;
  FPanel.SetFocus;
  DoDroppedDown;
end;

function TCustomTntExDropDownEdit.GetDropDownStyle: TTntExDropDownStyle;
begin
  result := csDropDown;
end;

function TCustomTntExDropDownEdit.IsEditable: boolean;
begin
  result := (FDropDownStyle <> csDropDownList);
end;

procedure TCustomTntExDropDownEdit.WMLButtonDown(var Message: TWMLButtonDown);
begin
  if IsEditable then
    inherited
  else
    NonEditMouseDown(Message);
end;

procedure TCustomTntExDropDownEdit.NonEditMouseDown(var Message: TWMLButtonDown);
var
  OnButton, WasFocused : boolean;
  p : TPoint;
begin
  WasFocused := Focused;
  p := point(message.XPos,message.YPos);
  OnButton := HitButton(p.x,p.y);

  SetFocus;
  HideCaret (Handle);

  if WasFocused or OnButton then
  begin
    if FPanel.Visible then
      CloseUp
    else
      DropDown;
  end;
end;

procedure TCustomTntExDropDownEdit.WMSetFocus(var Message: TWMSetFocus);
begin
  if not IsEditable then
    HideCaret(Handle)
  else
    ShowCaret(Handle);
  SelectAll;
  inherited;
end;

procedure TCustomTntExDropDownEdit.KeyPress(var Key: Char);
begin
  case Key of
    ^H, ^V, ^X, #32..#255:
    begin
      if IsEditable then
        Inherited
      else
        Key := #0;
    end;
  else
    inherited;
  end;
end;

procedure TCustomTntExDropDownEdit.DoCustomPaint;
begin

end;

function TCustomTntExDropDownEdit.GetCanvas: TCanvas;
begin
  result := FCanvas;
end;

end.
