{*****************************************************************************}
{                                                                             }
{    Tnt Extended Editors                                                     }
{      http://lummie.co.uk                                                    }
{        Version: 1.0.3                                                       }
{                                                                             }
{    Copyright (c) 2005 Matt harrison (http://www.lummie.co.uk)               }
{    Descended from TNT Controls                                              }
{        Copyright (c) 2002, 2003 Troy Wolbrink (troy.wolbrink@ccci.org)      }
{    VirtualTrees (c) Mike Lischke (public@soft-gems.net, www.soft-gems.net)  }
{                                                                             }
{*****************************************************************************}
unit TntExDropDownVirtualDrawTree;

interface


// remove remark to compile with VirtualTrees.
// Please ensure changes are reflected in TntExEditReg as well.
{$DEFINE VirtualTrees}

{$IFDEF VirtualTrees}
uses
  SysUtils, Classes, Controls, StdCtrls, TntStdCtrls, TntExDropDownControl, Messages, Forms, windows, VirtualTrees, graphics;

type
  TCustomTntExDropDownVirtualDrawTree = class;

  TTntExVirtualDrawTreeWrapper = class(TVirtualDrawTree)
  private
    FOwner : TCustomTntExDropDownVirtualDrawTree;
  protected
    procedure DoChange(Node: PVirtualNode); override;
    procedure DoExpanded(Node: PVirtualNode); override;
    procedure DoCollapsed(Node: PVirtualNode); override;
  public
    constructor Create(AOwner: TComponent); override;
    function FocusedNodeText : widestring;
    function FocusedImageIndex : integer;
    procedure PaintFocusedImage(Canvas : TCanvas; x,y : integer; IsEnabled : boolean);
    procedure PaintFocusedLines(Canvas : TCanvas; x,y : integer; IsEnabled : boolean);
  end;


  TCustomTntExDropDownVirtualDrawTree = class(TCustomTntExDropDownControl)
  private
    { Private declarations }
    FTree : TTntExVirtualDrawTreeWrapper;
    FOnChange: TNotifyEvent;
    procedure TreeKeyDown(Sender : TObject; var Key: Word; Shift: TShiftState);
    procedure SetTree(const Value: TVirtualDrawTree);
    function GetTree: TVirtualDrawTree;
  protected
    { Protected declarations }
    function DoPopulatePanel(APanel: TTntExDropDownPanel): Boolean; override;
    procedure DoDroppedDown; override;
    procedure DoClosedUp; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure Paint; override;
    procedure DoChange; virtual;
    procedure CMEnabledchanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Tree : TVirtualDrawTree read GetTree write SetTree;
    property PanelAutoWidth;
    property PanelWidth;
    property PanelHeight;
    property OnDroppedDown;
    property OnClosedUp;
    property DropDownPopupMenu;
    property OnChange : TNotifyEvent read FOnChange write FOnChange;
  end;

  TTntExDropDownVirtualDrawTree = class(TCustomTntExDropDownVirtualDrawTree)
  published
    { Published declarations }
    property Tree;
    property PanelAutoWidth;
    property PanelWidth;
    property PanelHeight;
    property OnDroppedDown;
    property OnClosedUp;
    property OnChange;
    property DropDownPopupMenu;
  end;

{$ENDIF}
implementation

uses ImgList;
{$IFDEF VirtualTrees}

const
  TreeExpandButtonIndent = 16;

{ TTntExVirtualDrawTreeWrapper }

constructor TTntExVirtualDrawTreeWrapper.Create(AOwner: TComponent);
begin
  inherited;
  FOwner := nil;
  assert(AOwner is TCustomTntExDropDownVirtualDrawTree);
  FOwner := TCustomTntExDropDownVirtualDrawTree(AOwner);
end;

procedure TTntExVirtualDrawTreeWrapper.DoChange(Node: PVirtualNode);
begin
  inherited;
  FOwner.Invalidate;
  FOwner.DoChange;
end;

procedure TTntExVirtualDrawTreeWrapper.DoCollapsed(Node: PVirtualNode);
begin
  inherited;
  FOwner.Invalidate;
end;

procedure TTntExVirtualDrawTreeWrapper.DoExpanded(Node: PVirtualNode);
begin
  inherited;
  FOwner.Invalidate;
end;

function TTntExVirtualDrawTreeWrapper.FocusedImageIndex: integer;
var
  AImageIndex : integer;
  Ghosted : Boolean;
begin
  AImageIndex := -1;
  if assigned(FocusedNode) then
  begin
    DoGetImageIndex(FocusedNode,ikNormal,Header.MainColumn,Ghosted,AImageIndex);
  end;
  result := AImageIndex;
end;

function TTntExVirtualDrawTreeWrapper.FocusedNodeText: widestring;
begin
  result := '';
end;

procedure TTntExVirtualDrawTreeWrapper.PaintFocusedImage(Canvas: TCanvas;
  x, y: integer; IsEnabled : boolean);
var
  ImageIndex : integer;
begin
  ImageIndex := FocusedImageIndex;
  if assigned(Images) and (ImageIndex <> -1) then
  begin
    if (vsHasChildren in FocusedNode.States) then x := x + TreeExpandButtonIndent;
    Images.Draw(Canvas,x,y,ImageIndex,IsEnabled);
  end;
end;

procedure TTntExVirtualDrawTreeWrapper.PaintFocusedLines(Canvas : TCanvas; x,y : integer; IsEnabled : boolean);
var
  Node : PVirtualNode;
  CellRect : TRect;
  ButtonX, ButtonY : integer;
begin
  Node := FocusedNode;
  if assigned(Node) then
  begin
    CellRect := Rect(0,0,16,16);
    ButtonX := x+2;
    ButtonY := y+4;
    // Show node button if allowed, if there child nodes and at least one of the child
    // nodes is visible or auto button hiding is disabled.
    if (toShowButtons in TreeOptions.PaintOptions) and (vsHasChildren in Node.States) and
      not ((vsAllChildrenHidden in Node.States) and
      (toAutoHideButtons in TreeOptions.AutoOptions)) then
      PaintNodeButton(Canvas, Node, CellRect, ButtonX, ButtonY, BidiMode);
  end;
end;


{ TntExDropDownVirtualDrawTree }

constructor TCustomTntExDropDownVirtualDrawTree.Create(AOwner: TComponent);
begin
  inherited;
  FOnChange := nil;
  FTree := TTntExVirtualDrawTreeWrapper.Create(self);
  FTree.OnKeyDown := TreeKeyDown;
  FTree.Name := 'DropDownTree';
  FTree.visible := false;
  FTree.parent := self.Panel;
  FTree.SetSubComponent(true);
  TabStop := true;
end;

destructor TCustomTntExDropDownVirtualDrawTree.Destroy;
begin
  inherited;
end;

procedure TCustomTntExDropDownVirtualDrawTree.DoClosedUp;
begin
  inherited;
end;

procedure TCustomTntExDropDownVirtualDrawTree.DoDroppedDown;
begin
  inherited;
  FTree.SetFocus;
end;

function TCustomTntExDropDownVirtualDrawTree.DoPopulatePanel(
  APanel: TTntExDropDownPanel): Boolean;
begin
  result := true;
  FTree.align := alclient;
  FTree.parent := APanel;
  FTree.visible := true;
end;

procedure TCustomTntExDropDownVirtualDrawTree.TreeKeyDown(Sender : TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Shift = []) then
  begin
    case key of
      VK_Escape, VK_Return : KeyDown(Key,Shift);
    end;
  end;
  if (ssAlt in Shift) and (Key = VK_Down) then KeyDown(Key,Shift);
end;


procedure TCustomTntExDropDownVirtualDrawTree.SetTree(
  const Value: TVirtualDrawTree);
begin
  FTree.Assign(Value);
end;

function TCustomTntExDropDownVirtualDrawTree.GetTree: TVirtualDrawTree;
begin
  result := FTree;
end;

procedure TCustomTntExDropDownVirtualDrawTree.KeyDown(var Key: Word;
  Shift: TShiftState);
var
  NewNode : PVirtualNode;
  handled : boolean;
begin
  handled := false;
  NewNode := FTree.FocusedNode;
  if (Shift = []) and assigned(NewNode) then
  begin
    handled := true;
    case key of
      VK_DOWN : begin
        if not assigned(NewNode) then
          NewNode := FTree.GetFirst
        else
          NewNode := FTree.GetNextVisible(NewNode);
      end;
      VK_UP : begin
        if not assigned(NewNode) then
          NewNode := FTree.GetLast
        else
          NewNode := FTree.GetPreviousVisible(NewNode);
      end;
      VK_Right : begin
        if vsHasChildren in NewNode.States then
          NewNode := FTree.GetFirstChild(NewNode);
      end;
      VK_Left : begin
        if NewNode.Parent <> Ftree.RootNode then
          NewNode := NewNode.Parent;
      end;
      VK_Home : begin
          NewNode := FTree.GetFirst;
      end;
      VK_END : begin
          NewNode := FTree.GetLast;
      end;
    else
      handled := false;
    end;
  end;

  if handled then
  begin
    if not assigned(NewNode) then NewNode := Ftree.FocusedNode;
    FTree.ClearSelection;
    FTree.FocusedNode := NewNode;
    if assigned(NewNode) then FTree.selected[newNode] := true;
    FTree.DoChange(NewNode);
  end
  else
    inherited;
end;


procedure TCustomTntExDropDownVirtualDrawTree.WMChar(var Message: TWMChar);
var
  AMessage : TMessage;
begin
  AMessage := TMessage(Message);
  PostMessage(FTree.Handle, AMessage.Msg,AMessage.WParam,AMessage.LParam);
  inherited;
end;

procedure TCustomTntExDropDownVirtualDrawTree.WMKeyDown(
  var Message: TWMKeyDown);
var
  AMessage : TMessage;
begin
  if not DoKeyDown(Message) then
  begin
    AMessage := TMessage(Message);
    PostMessage(FTree.Handle, AMessage.Msg,AMessage.WParam,AMessage.LParam);
  end;
end;

procedure TCustomTntExDropDownVirtualDrawTree.Paint;
begin
  inherited;
end;

{$ENDIF}
procedure TCustomTntExDropDownVirtualDrawTree.DoChange;
begin
  if assigned(FOnChange) then FOnChange(self);
end;

procedure TCustomTntExDropDownVirtualDrawTree.CMEnabledchanged(
  var Message: TMessage);
begin
  inherited;
  invalidate;
end;

end.
