{*****************************************************************************}
{                                                                             }
{    Tnt Extended Editors                                                     }
{      http://lummie.co.uk                                                    }
{        Version: 1.0.3                                                       }
{                                                                             }
{    Copyright (c) 2005 Matt harrison (http://www.lummie.co.uk)               }
{    Descended from TNT Controls                                              }
{        Copyright (c) 2002, 2003 Troy Wolbrink (troy.wolbrink@ccci.org)      }
{    VirtualTrees (c) Mike Lischke (public@soft-gems.net, www.soft-gems.net)  }
{                                                                             }
{*****************************************************************************}
{
    Version 1.0.3
      Added GetTreeClass
      Fixed moving two nodes at a time when up or down pressed when the tree was closed up
      Added Mouse Wheel support
}

unit TntExDropDownVirtualStringTree;

interface


// remove remark to compile with VirtualTrees.
// Please ensure changes are reflected in TntExEditReg as well.
{$DEFINE VirtualTrees}

{$IFDEF VirtualTrees}
uses
  SysUtils, Classes, Controls, StdCtrls, TntStdCtrls, TntExDropDownEdit, Messages, Forms, windows, VirtualTrees, graphics;

type
  TTntExDropDownVirtualStringTree = class;

  TTntExVirtualStringTreeWrapper = class(TVirtualStringTree)
  private
    FOwner : TTntExDropDownVirtualStringTree;
  protected
    procedure DoChange(Node: PVirtualNode); override;
    procedure DoExpanded(Node: PVirtualNode); override;
    procedure DoCollapsed(Node: PVirtualNode); override;
  public
    constructor Create(AOwner: TComponent); override;
    function FocusedNodeText : widestring;
    function FocusedImageIndex : integer;
    procedure PaintFocusedImage(Canvas : TCanvas; x,y : integer; IsEnabled : boolean);
    procedure PaintFocusedLines(Canvas : TCanvas; x,y : integer; IsEnabled : boolean);
  end;


  TTntExDropDownVirtualStringTree = class(TCustomTntExDropDownEdit)
  private
    { Private declarations }
    FTree : TTntExVirtualStringTreeWrapper;
    procedure TreeKeyDown(Sender : TObject; var Key: Word; Shift: TShiftState);
    procedure SetTree(const Value: TVirtualStringTree);
    function GetTree: TVirtualStringTree;
  protected
    { Protected declarations }
    function DoPopulatePanel(APanel: TTntExDropDownPanel): Boolean; override;
    procedure DoDroppedDown; override;
    procedure DoClosedUp; override;
    procedure Loaded; override;
    procedure UpdateEditFromTree;
    procedure UpdateMargins; override;
    procedure DoCustomPaint; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    function GetDropDownStyle: TTntExDropDownStyle; override;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    function GetTreeClass(AOwner : TComponent) : TTntExVirtualStringTreeWrapper;  virtual;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property Tree : TVirtualStringTree read GetTree write SetTree;
    property PanelAutoWidth;
    property PanelWidth;
    property PanelHeight;
    property OnDroppedDown;
    property OnClosedUp;
    property DropDownPopupMenu;
  end;

{$ENDIF}
implementation

uses ImgList;
{$IFDEF VirtualTrees}

const
  TreeExpandButtonIndent = 16;

{ TTntExVirtualStringTreeWrapper }

constructor TTntExVirtualStringTreeWrapper.Create(AOwner: TComponent);
begin
  inherited;
  FOwner := nil;
  if AOwner is TTntExDropDownVirtualStringTree then
    FOwner := TTntExDropDownVirtualStringTree(AOwner);
end;

procedure TTntExVirtualStringTreeWrapper.DoChange(Node: PVirtualNode);
begin
  inherited;
  FOwner.UpdateMargins;
  FOwner.UpdateEditFromTree;
  FOwner.Invalidate;
end;

procedure TTntExVirtualStringTreeWrapper.DoCollapsed(Node: PVirtualNode);
begin
  inherited;
  FOwner.Invalidate;
end;

procedure TTntExVirtualStringTreeWrapper.DoExpanded(Node: PVirtualNode);
begin
  inherited;
  FOwner.Invalidate;
end;

function TTntExVirtualStringTreeWrapper.FocusedImageIndex: integer;
var
  AImageIndex : integer;
  Ghosted : Boolean;
begin
  AImageIndex := -1;
  if assigned(FocusedNode) then
  begin
    DoGetImageIndex(FocusedNode,ikNormal,Header.MainColumn,Ghosted,AImageIndex);
  end;
  result := AImageIndex;
end;

function TTntExVirtualStringTreeWrapper.FocusedNodeText: widestring;
var
  AText : widestring;
begin
  AText := '';
  if assigned(FocusedNode) then
  begin
    DoGetText(FocusedNode,Header.MainColumn,ttNormal,AText);
  end;
  result := AText;
end;

procedure TTntExVirtualStringTreeWrapper.PaintFocusedImage(Canvas: TCanvas;
  x, y: integer; IsEnabled : boolean);
var
  ImageIndex : integer;
begin
  ImageIndex := FocusedImageIndex;
  if assigned(Images) and (ImageIndex <> -1) then
  begin
    if (vsHasChildren in FocusedNode.States) then x := x + TreeExpandButtonIndent;
    Images.Draw(Canvas,x,y,ImageIndex,IsEnabled);
  end;
end;

procedure TTntExVirtualStringTreeWrapper.PaintFocusedLines(Canvas : TCanvas; x,y : integer; IsEnabled : boolean);
var
  Node : PVirtualNode;
  CellRect : TRect;
  ButtonX, ButtonY : integer;
begin
  Node := FocusedNode;
  if assigned(Node) then
  begin
    CellRect := Rect(0,0,16,16);
    ButtonX := x+2;
    ButtonY := y+4;
    // Show node button if allowed, if there child nodes and at least one of the child
    // nodes is visible or auto button hiding is disabled.
    if (toShowButtons in TreeOptions.PaintOptions) and (vsHasChildren in Node.States) and
      not ((vsAllChildrenHidden in Node.States) and
      (toAutoHideButtons in TreeOptions.AutoOptions)) then
      PaintNodeButton(Canvas, Node, CellRect, ButtonX, ButtonY, BidiMode);
  end;
end;


{ TntExDropDownVirtualStringTree }

constructor TTntExDropDownVirtualStringTree.Create(AOwner: TComponent);
begin
  inherited;
  FTree := GetTreeClass(Self);
  FTree.OnKeyDown := TreeKeyDown;
  FTree.Name := 'DropDownTree';
  FTree.visible := false;
  FTree.parent := self.Panel;
  FTree.SetSubComponent(true);
end;

destructor TTntExDropDownVirtualStringTree.Destroy;
begin
  inherited;
end;

procedure TTntExDropDownVirtualStringTree.DoClosedUp;
begin
  inherited;
  SelectAll;
end;

procedure TTntExDropDownVirtualStringTree.DoDroppedDown;
begin
  inherited;
  FTree.SetFocus;
end;

function TTntExDropDownVirtualStringTree.DoPopulatePanel(
  APanel: TTntExDropDownPanel): Boolean;
begin
  result := true;
  FTree.align := alclient;
  FTree.parent := APanel;
  FTree.visible := true;
end;

procedure TTntExDropDownVirtualStringTree.TreeKeyDown(Sender : TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Shift = []) then
  begin
    case key of
      VK_Escape, VK_Return : KeyDown(Key,Shift);
    end;
  end;
  if (ssAlt in Shift) and (Key = VK_Down) then KeyDown(Key,Shift);
end;


procedure TTntExDropDownVirtualStringTree.SetTree(
  const Value: TVirtualStringTree);
begin
  FTree.Assign(Value);
end;


procedure TTntExDropDownVirtualStringTree.Loaded;
begin
  inherited;
  UpdateEditFromTree;
end;

function TTntExDropDownVirtualStringTree.GetTree: TVirtualStringTree;
begin
  result := FTree;
end;

procedure TTntExDropDownVirtualStringTree.UpdateEditFromTree;
begin
  Text := FTree.FocusedNodeText;
end;

procedure TTntExDropDownVirtualStringTree.UpdateMargins;
var
  IWidth : integer;
begin
  inherited;
  IWidth := 0;

  if assigned(FTree.Images) and (FTree.FocusedImageIndex <> -1) then
  begin
    IWidth := FTree.Images.Width + 4;
  end;
  if assigned(FTree.FocusedNode) and (vsHasChildren in FTree.FocusedNode.States) then
  begin
    IWidth := IWidth + TreeExpandButtonIndent;
  end;

  if IWidth <> 0 then
  begin
    Perform(EM_SETMARGINS,EC_LEFTMARGIN,IWidth);
  end;
end;

procedure TTntExDropDownVirtualStringTree.DoCustomPaint;
begin
  inherited;
  if Canvas.HandleAllocated then
  begin
    FTree.PaintFocusedImage(Canvas,2,0,enabled);
    FTree.PaintFocusedLines(Canvas,2,0,enabled);
  end;
end;

procedure TTntExDropDownVirtualStringTree.KeyDown(var Key: Word;
  Shift: TShiftState);
var
  NewNode : PVirtualNode;
  handled : boolean;
begin
  handled := false;
  NewNode := FTree.FocusedNode;
  if (Shift = []) and assigned(NewNode) then
  begin
    handled := true;
    case key of
      VK_DOWN : begin
        if not assigned(NewNode) then
          NewNode := FTree.GetFirst
      end;
      VK_UP : begin
        if not assigned(NewNode) then
          NewNode := FTree.GetLast
      end;
      VK_Right : begin
        if vsHasChildren in NewNode.States then
          NewNode := FTree.GetFirstChild(NewNode);
      end;
      VK_Left : begin
        if NewNode.Parent <> Ftree.RootNode then
          NewNode := NewNode.Parent;
      end;
      VK_Home : begin
          NewNode := FTree.GetFirst;
      end;
      VK_END : begin
          NewNode := FTree.GetLast;
      end;

    else
      handled := false;
    end;
  end;

  if handled then
  begin
    if not assigned(NewNode) then NewNode := Ftree.FocusedNode;
    FTree.ClearSelection;
    FTree.FocusedNode := NewNode;
    if assigned(NewNode) then FTree.selected[newNode] := true;
    FTree.DoChange(NewNode);
  end
  else
    inherited;
end;

function TTntExDropDownVirtualStringTree.GetDropDownStyle: TTntExDropDownStyle;
begin
  result := csDropDownList;
end;


procedure TTntExDropDownVirtualStringTree.WMChar(var Message: TWMChar);
var
  AMessage : TMessage;
begin
  AMessage := TMessage(Message);
  PostMessage(FTree.Handle, AMessage.Msg,AMessage.WParam,AMessage.LParam);
  inherited;
end;

procedure TTntExDropDownVirtualStringTree.WMKeyDown(
  var Message: TWMKeyDown);
var
  AMessage : TMessage;
begin
  if not DoKeyDown(Message) then
  begin
    AMessage := TMessage(Message);
    PostMessage(FTree.Handle, AMessage.Msg,AMessage.WParam,AMessage.LParam);
  end;
end;

{$ENDIF}
function TTntExDropDownVirtualStringTree.GetTreeClass(AOwner : TComponent): TTntExVirtualStringTreeWrapper;
begin
  result := TTntExVirtualStringTreeWrapper.create(AOwner);
end;


function TTntExDropDownVirtualStringTree.DoMouseWheel(Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint): Boolean;
var
  NewNode : PVirtualNode;
begin
  result := true;
  NewNode := FTree.FocusedNode;
  if WheelDelta < 0 then
  begin
    if not assigned(NewNode) then
       NewNode := FTree.GetFirst
    else
      newNode := FTree.GetNextSibling(NewNode);
  end
  else
  begin
    if not assigned(NewNode) then
       NewNode := FTree.GetLast
    else
      newNode := FTree.GetPreviousSibling(NewNode);
  end;

  if assigned(NewNode) then
  begin
    FTree.ClearSelection;
    FTree.FocusedNode := NewNode;
    if assigned(NewNode) then FTree.selected[newNode] := true;
    FTree.DoChange(NewNode);
    SelectAll;
  end;
end;

end.
