{*****************************************************************************}
{                                                                             }
{    Tnt Extended Editors                                                     }
{      http://lummie.co.uk                                                    }
{        Version: 1.0.3                                                      }
{                                                                             }
{    Copyright (c) 2005 Matt harrison (http://www.lummie.co.uk)               }
{    Descended from TNT Controls                                              }
{        Copyright (c) 2002, 2003 Troy Wolbrink (troy.wolbrink@ccci.org)      }                                                                   
{                                                                             }
{*****************************************************************************}
unit TntExWinControl;

interface

uses
  SysUtils, Classes, Controls, tntControls, messages, TntActnList, tntForms, tntClasses, windows,
  graphics, themes, extctrls, forms;

type
  TTntExWinControl = class(TCustomControl)
  private
    { Private declarations }
    function GetHint: WideString;
    procedure SetHint(const Value: WideString);
    function IsHintStored: Boolean;
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
  protected
    { Protected declarations }
    procedure DefineProperties(Filer: TFiler); override;
    function GetActionLinkClass: TControlActionLinkClass; override;
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
    property Hint: WideString read GetHint write SetHint stored IsHintStored;
    procedure Paint; override;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
  published
    { Published declarations }
  end;

implementation

{ TTntExWinControl }

procedure TTntExWinControl.ActionChange(Sender: TObject;
  CheckDefaults: Boolean);
begin
  TntControl_BeforeInherited_ActionChange(Self, Sender, CheckDefaults);
  inherited;
end;

procedure TTntExWinControl.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsWideCharAccel(CharCode, Caption) and Enabled and Visible and
      (Parent <> nil) and Parent.Showing then
    begin
      Click;
      Result := 1;
    end else
      inherited;
end;

procedure TTntExWinControl.CMHintShow(var Message: TMessage);
begin
  ProcessCMHintShowMsg(Message);
  inherited;
end;

constructor TTntExWinControl.Create(AOwner: TComponent);
begin
  inherited;
  BorderWidth := 1;
end;

procedure TTntExWinControl.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    if NewStyleControls and Ctl3D then
    begin
      Style := Style and not WS_BORDER or WS_TABSTOP;
      if not ThemeServices.ThemesEnabled then ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TTntExWinControl.DefineProperties(Filer: TFiler);
begin
  inherited;
  TntPersistent_AfterInherited_DefineProperties(Filer, Self);
end;

function TTntExWinControl.GetActionLinkClass: TControlActionLinkClass;
begin
  Result := TntControl_GetActionLinkClass(Self, inherited GetActionLinkClass);
end;

function TTntExWinControl.GetHint: WideString;
begin
  Result := TntControl_GetHint(Self)
end;

function TTntExWinControl.IsHintStored: Boolean;
begin
  Result := TntControl_IsHintStored(Self)
end;

procedure TTntExWinControl.Paint;
var
  ARect : TRect;
  State : TThemedEdit;
  Details: TThemedElementDetails;
begin
  inherited;
  canvas.Brush.Style := bsSolid;
  canvas.Brush.Color := clWindow;
  canvas.FillRect(rect(0,0,width,height));
  if ThemeServices.ThemesEnabled then
  begin
    State := teEditTextNormal;
    if not enabled then State := teEditTextDisabled;
    Details := ThemeServices.GetElementDetails(State);
    ARect := Rect(0,0,width-2,height-2);
    ThemeServices.DrawElement(Canvas.Handle, Details, ARect);
  end
end;

procedure TTntExWinControl.SetHint(const Value: WideString);
begin
  TntControl_SetHint(Self, Value);
end;

end.
